package org.pathwaycommons.trans.util;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeMap;
import org.genemania.dw.util.GenUtil;
import org.jdom.Element;
import org.jdom.Namespace;
import org.pathwaycommons.trans.PSIMITags;

/**
 * Singleton class for simple interaction stats.
 * interactionsMap: key: interaction source ID; Value: auto-generated ID
 * intAttMap: key: interaction primary reference ID; interaction source ID(s)
 *
 * This class has resorted to using custom tab delimiters with some string
 * manipulations. If need arises this data may evolve into its own separate
 * data structure.
 *
 * @author rashadbadrawi
 */

public class TransStats {

    private TreeMap  <String, ArrayList <String>> interactionsMap = new TreeMap <String, ArrayList <String>> ();
    private TreeMap <String, ArrayList <String>> intAttMap = new TreeMap <String, ArrayList <String>> ();
    private TreeMap <String, ArrayList <String>> intRefMap = new TreeMap <String, ArrayList <String>> ();
    private TreeMap <String, ArrayList <String>> refIntMap = new TreeMap <String, ArrayList <String>> ();
    private ArrayList <String> processedInteractionsList = new ArrayList <String> ();
    private ArrayList <String> processedInteractorsList = new ArrayList <String> ();

    private static TransStats statsInstance;

    public static TransStats getTransStats () {

        if (statsInstance == null) {
            statsInstance = new TransStats ();
        }

        return statsInstance;
    }

    private TransStats () {}

    protected void trackIntPSIMI (ArrayList <Element> interactionsList) {

        Namespace psins = Namespace.getNamespace(PSIMITags.PSIMI_NS);
        for (int i = 0; i < interactionsList.size (); i++) {
            Element interactionE = interactionsList.get (i);
            //interaction ID
            Element xrefE = interactionE.getChild (PSIMITags.XREF, psins);
            String interactionID = xrefE.getChild (PSIMITags.PREF, psins)
                                        .getAttributeValue(PSIMITags.ID_ATT);
            trackIntElement (this.interactionsMap,
                             xrefE.getChild (PSIMITags.PREF, psins)
                                  .getAttributeValue(PSIMITags.DB_ATT),
                             interactionID,
                             interactionE.getAttributeValue (PSIMITags.ID_ATT));
            //pubmed ID
            ArrayList <Element> expDescList = new ArrayList <Element> ();
            expDescList.addAll (interactionE.getChild(PSIMITags.EXP_LIST, psins)
                                            .getChildren (PSIMITags.EXP_DESC, psins));
            for (int j = 0; j < expDescList.size (); j++) {
                Element prefE = expDescList.get (j).getChild (PSIMITags.BIB, psins)
                                                   .getChild (PSIMITags.XREF, psins)
                                                   .getChild (PSIMITags.PREF, psins);
                trackIntElement (this.intRefMap,
                                 prefE.getAttributeValue(PSIMITags.DB_ATT),
                                 prefE.getAttributeValue(PSIMITags.ID_ATT),
                                 interactionID);
            }
            //interactor ID
            ArrayList <Element> partList = new ArrayList <Element> ();
            partList.addAll (interactionE.getChild (PSIMITags.PARTLIST, psins)
                                         .getChildren (PSIMITags.PART, psins));
            for (int j = 0; j < partList.size(); j++) {
                Element interactorE = partList.get (j).getChild (PSIMITags.INTERACTOR, psins);
                if (interactorE == null) {
                    System.err.println ("No interactor. InteractorRef not supported: " +
                                         partList.get (j));
                    return;
                }
                xrefE = interactorE.getChild (PSIMITags.XREF, psins);
                if (xrefE != null) {
                    trackIntElement (this.intAttMap,
                                     xrefE.getChild (PSIMITags.PREF, psins)
                                          .getAttributeValue(PSIMITags.DB_ATT),
                                     xrefE.getChild (PSIMITags.PREF, psins)
                                          .getAttributeValue(PSIMITags.ID_ATT),
                                     interactionID);
                }
            }
        }
    }

    protected boolean isRedundant (Element element) {

        Namespace psins = Namespace.getNamespace(PSIMITags.PSIMI_NS);
        Element xrefE = element.getChild (PSIMITags.XREF, psins);
        if (xrefE != null) {
            String ID = xrefE.getChild(PSIMITags.PREF, psins).getAttributeValue(PSIMITags.ID_ATT);
            String IDType = xrefE.getChild(PSIMITags.PREF, psins).getAttributeValue(PSIMITags.DB_ATT);
            if (element.getName().equals(PSIMITags.INT)) {
                return isRedundantInteraction(IDType, ID);
            } else { //if (element.getName ().equals (PSIMITags.INTERACTOR)) {
                return isRedundantInteractor(IDType, ID);
            }
        } else {
            return false;
        }
    }

    private boolean isRedundantInteraction (String IDType, String ID) {

       if (!this.processedInteractionsList.contains (IDType + GenUtil.AT + ID)) {
          this.processedInteractionsList.add (IDType + GenUtil.AT + ID);
          return false;
       }
       System.err.println ("Error: Redundant Interaction.");

       return true;
    }

    private boolean isRedundantInteractor (String IDType, String ID) {

       if (!this.processedInteractorsList.contains (IDType + GenUtil.AT + ID)) {
          this.processedInteractorsList.add (IDType + GenUtil.AT + ID);
          return false;
       }

       return true;
    }

    private void trackIntElement (TreeMap <String, ArrayList <String>> tempMap,
                               String IDType, String ID, String interactionID) {

        ArrayList <String> tempList;
        if (!tempMap.containsKey (IDType + GenUtil.AT + ID)) {
            tempList = new ArrayList <String> ();
            tempList.add (interactionID);
            tempMap.put (IDType + GenUtil.AT + ID, tempList);
        } else {
            tempList = tempMap.get (IDType + GenUtil.AT + ID);
            if (!tempList.contains (interactionID)) {
                tempList.add (interactionID);
                tempMap.put (IDType + GenUtil.AT + ID, tempList);
            }
        }
    }

    protected void dumpStats () {

        System.out.println ("Dumping interactions list: " + interactionsMap.size());
        Iterator iterator = interactionsMap.keySet().iterator();
        while (iterator.hasNext()) {
            String currentInteractionID = (String)iterator.next ();
            System.out.println ("Interaction (BIND ID vs. PSI-MI ID): "      + GenUtil.TAB +
                                currentInteractionID + GenUtil.TAB +
            interactionsMap.get (currentInteractionID).size () + GenUtil.TAB +
                                 GenUtil.addList (new String (),
                                 interactionsMap.get (currentInteractionID)));
        }

        System.out.println ("Dumping interactants list (by pRef)" + intAttMap.size());
        iterator = intAttMap.keySet().iterator();
        while (iterator.hasNext()) {
            String currentAttID = (String)iterator.next ();
            System.out.println ("Interactant vs. Interaction: " + GenUtil.TAB +
                                currentAttID    + GenUtil.TAB +
            intAttMap.get (currentAttID).size ()+ GenUtil.TAB +
                                GenUtil.addList (new String (),
                                intAttMap.get (currentAttID)));
        }

        System.out.println ("Dumping references list" + intRefMap.size());
        iterator = intRefMap.keySet().iterator();
        while (iterator.hasNext()) {
            String currentIntRefID = (String)iterator.next ();
            System.out.println ("Pubmed vs. Interaction: " + GenUtil.TAB +
                                currentIntRefID + GenUtil.TAB +
            intRefMap.get (currentIntRefID).size () + GenUtil.TAB +
                                GenUtil.addList (new String (), 
                                intRefMap.get (currentIntRefID)));
            //process reverse pubmed.
            ArrayList <String> tempList = intRefMap.get (currentIntRefID);
            for (int i = 0; i < tempList.size (); i++) {
                String currentInteractionID = tempList.get (i);
                this.trackIntElement(this.refIntMap, PSIMITags.INT,
                                     currentInteractionID,
                                     currentIntRefID);
            }
        }
        //dump reverse pubmed
        System.out.println ("Dumping references list 2" + refIntMap.size());
        iterator = refIntMap.keySet().iterator();
        while (iterator.hasNext()) {
            String currentRefIntMapID = (String)iterator.next ();
            System.out.println ("Interaction vs. Pubmed: " + GenUtil.TAB +
                                currentRefIntMapID + GenUtil.TAB +
            refIntMap.get (currentRefIntMapID).size () + GenUtil.TAB +
                                GenUtil.addList (new String (),
                                refIntMap.get (currentRefIntMapID)));
        }
    }
}
